@@ -2,45 +2,52 @@ package LWP::Protocol::http;
 
 use strict;
 
+our $VERSION = '6.61';
+
 require HTTP::Response;
 require HTTP::Status;
 require Net::HTTP;
 
-use vars qw(@ISA @EXTRA_SOCK_OPTS);
-
-require LWP::Protocol;
-@ISA = qw(LWP::Protocol);
+use parent qw(LWP::Protocol);
 
+our @EXTRA_SOCK_OPTS;
 my $CRLF = "\015\012";
 
 sub _new_socket
 {
     my($self, $host, $port, $timeout) = @_;
-    my $conn_cache = $self->{ua}{conn_cache};
-    if ($conn_cache) {
-	if (my $sock = $conn_cache->withdraw("http", "$host:$port")) {
-	    return $sock if $sock && !$sock->can_read(0);
-	    # if the socket is readable, then either the peer has closed the
-	    # connection or there are some garbage bytes on it.  In either
-	    # case we abandon it.
-	    $sock->close;
-	}
+
+    # IPv6 literal IP address should be [bracketed] to remove
+    # ambiguity between ip address and port number.
+    if ( ($host =~ /:/) && ($host !~ /^\[/) ) {
+      $host = "[$host]";
     }
 
     local($^W) = 0;  # IO::Socket::INET can be noisy
     my $sock = $self->socket_class->new(PeerAddr => $host,
 					PeerPort => $port,
+					LocalAddr => $self->{ua}{local_address},
 					Proto    => 'tcp',
 					Timeout  => $timeout,
-					KeepAlive => !!$conn_cache,
-					SendTE    => 1,
+					KeepAlive => !!$self->{ua}{conn_cache},
+					SendTE    => $self->{ua}{send_te},
 					$self->_extra_sock_opts($host, $port),
 				       );
 
     unless ($sock) {
 	# IO::Socket::INET leaves additional error messages in $@
-	$@ =~ s/^.*?: //;
-	die "Can't connect to $host:$port ($@)";
+	my $status = "Can't connect to $host:$port";
+	if ($@ =~ /\bconnect: (.*)/ ||
+	    $@ =~ /\b(Bad hostname)\b/ ||
+	    $@ =~ /\b(nodename nor servname provided, or not known)\b/ ||
+	    $@ =~ /\b(certificate verify failed)\b/ ||
+	    $@ =~ /\b(Crypt-SSLeay can't verify hostnames)\b/
+	) {
+	    $status .= " ($1)";
+	} elsif ($@) {
+	    $status .= " ($@)";
+	}
+	die "$status\n\n$@";
     }
 
     # perl 5.005's IO::Socket does not have the blocking method.
@@ -49,6 +56,11 @@ sub _new_socket
     $sock;
 }
 
+sub socket_type
+{
+    return "http";
+}
+
 sub socket_class
 {
     my $self = shift;
@@ -91,9 +103,10 @@ sub _fixup_header
     }
     $h->init_header('Host' => $hhost);
 
-    if ($proxy) {
+    if ($proxy && $url->scheme ne 'https') {
 	# Check the proxy URI's userinfo() for proxy credentials
-	# export http_proxy="http://proxyuser:proxypass@proxyhost:port"
+	# export http_proxy="http://proxyuser:proxypass@proxyhost:port".
+	# For https only the initial CONNECT requests needs authorization.
 	my $p_auth = $proxy->userinfo();
 	if(defined $p_auth) {
 	    require URI::Escape;
@@ -121,32 +134,99 @@ sub request
     # check method
     my $method = $request->method;
     unless ($method =~ /^[A-Za-z0-9_!\#\$%&\'*+\-.^\`|~]+$/) {  # HTTP token
-	return new HTTP::Response &HTTP::Status::RC_BAD_REQUEST,
+	return HTTP::Response->new( HTTP::Status::RC_BAD_REQUEST,
 				  'Library does not allow method ' .
-				  "$method for 'http:' URLs";
+				  "$method for 'http:' URLs");
     }
 
     my $url = $request->uri;
-    my($host, $port, $fullpath);
 
-    # Check if we're proxy'ing
-    if (defined $proxy) {
-	# $proxy is an URL to an HTTP server which will proxy this request
-	$host = $proxy->host;
-	$port = $proxy->port;
-	$fullpath = $method eq "CONNECT" ?
-                       ($url->host . ":" . $url->port) :
-                       $url->as_string;
+    # Proxying SSL with a http proxy needs issues a CONNECT request to build a
+    # tunnel and then upgrades the tunnel to SSL. But when doing keep-alive the
+    # https request does not need to be the first request in the connection, so
+    # we need to distinguish between
+    # - not yet connected (create socket and ssl upgrade)
+    # - connected but not inside ssl tunnel (ssl upgrade)
+    # - inside ssl tunnel to the target - once we are in the tunnel to the
+    #   target we cannot only reuse the tunnel for more https requests with the
+    #   same target
+
+    my $ssl_tunnel = $proxy && $url->scheme eq 'https'
+	&& $url->host.":".$url->port;
+
+    my ($host,$port) = $proxy
+	? ($proxy->host,$proxy->port)
+	: ($url->host,$url->port);
+    my $fullpath =
+	$method eq 'CONNECT' ? $url->host . ":" . $url->port :
+	$proxy && ! $ssl_tunnel ? $url->as_string :
+	do {
+	    my $path = $url->path_query;
+	    $path = "/$path" if $path !~m{^/};
+	    $path
+	};
+
+    my $socket;
+    my $conn_cache = $self->{ua}{conn_cache};
+    my $cache_key;
+    if ( $conn_cache ) {
+	$cache_key = "$host:$port";
+	# For https we reuse the socket immediately only if it has an established
+	# tunnel to the target. Otherwise a CONNECT request followed by an SSL
+	# upgrade need to be done first. The request itself might reuse an
+	# existing non-ssl connection to the proxy
+	$cache_key .= "!".$ssl_tunnel if $ssl_tunnel;
+	if ( $socket = $conn_cache->withdraw($self->socket_type,$cache_key)) {
+	    if ($socket->can_read(0)) {
+		# if the socket is readable, then either the peer has closed the
+		# connection or there are some garbage bytes on it.  In either
+		# case we abandon it.
+		$socket->close;
+		$socket = undef;
+	    } # else use $socket
+	    else {
+		$socket->timeout($timeout);
+	    }
+	}
     }
-    else {
-	$host = $url->host;
-	$port = $url->port;
-	$fullpath = $url->path_query;
-	$fullpath = "/$fullpath" unless $fullpath =~ m,^/,;
+
+    if ( ! $socket && $ssl_tunnel ) {
+	my $proto_https = LWP::Protocol::create('https',$self->{ua})
+	    or die "no support for scheme https found";
+
+	# only if ssl socket class is IO::Socket::SSL we can upgrade
+	# a plain socket to SSL. In case of Net::SSL we fall back to
+	# the old version
+	if ( my $upgrade_sub = $proto_https->can('_upgrade_sock')) {
+	    my $response = $self->request(
+		HTTP::Request->new('CONNECT',"http://$ssl_tunnel"),
+		$proxy,
+		undef,$size,$timeout
+	    );
+	    $response->is_success or die
+		"establishing SSL tunnel failed: ".$response->status_line;
+	    $socket = $upgrade_sub->($proto_https,
+		$response->{client_socket},$url)
+		or die "SSL upgrade failed: $@";
+	} else {
+	    $socket = $proto_https->_new_socket($url->host,$url->port,$timeout);
+	}
+    }
+
+    if ( ! $socket ) {
+	# connect to remote site w/o reusing established socket
+	$socket = $self->_new_socket($host, $port, $timeout );
+    }
+
+    my $http_version = "";
+    if (my $proto = $request->protocol) {
+	if ($proto =~ /^(?:HTTP\/)?(1.\d+)$/) {
+	    $http_version = $1;
+	    $socket->http_version($http_version);
+	    $socket->send_te(0) if $http_version eq "1.0";
+	}
     }
 
-    # connect to remote site
-    my $socket = $self->_new_socket($host, $port, $timeout);
     $self->_check_sock($request, $socket);
 
     my @h;
@@ -156,7 +236,7 @@ sub request
     $request_headers->scan(sub {
 			       my($k, $v) = @_;
 			       $k =~ s/^://;
-			       $v =~ s/\n/ /g;
+			       $v =~ tr/\n/ /;
 			       push(@h, $k, $v);
 			   });
 
@@ -208,7 +288,7 @@ sub request
             my $n = $socket->syswrite($req_buf, length($req_buf));
             unless (defined $n) {
                 redo WRITE if $!{EINTR};
-                if ($!{EAGAIN}) {
+                if ($!{EWOULDBLOCK} || $!{EAGAIN}) {
                     select(undef, undef, undef, 0.1);
                     redo WRITE;
                 }
@@ -231,7 +311,15 @@ sub request
 	my $eof;
 	my $wbuf;
 	my $woffset = 0;
-	if (ref($content_ref) eq 'CODE') {
+      INITIAL_READ:
+	if ($write_wait) {
+	    # skip filling $wbuf when waiting for 100-continue
+	    # because if the response is a redirect or auth required
+	    # the request will be cloned and there is no way
+	    # to reset the input stream
+	    # return here via the label after the 100-continue is read
+	}
+	elsif (ref($content_ref) eq 'CODE') {
 	    my $buf = &$content_ref();
 	    $buf = "" unless defined($buf);
 	    $buf = sprintf "%x%s%s%s", length($buf), $CRLF, $buf, $CRLF
@@ -254,7 +342,7 @@ sub request
 	vec($fbits, fileno($socket), 1) = 1;
 
       WRITE:
-	while ($woffset < length($$wbuf)) {
+	while ($write_wait || $woffset < length($$wbuf)) {
 
 	    my $sel_timeout = $timeout;
 	    if ($write_wait) {
@@ -270,7 +358,7 @@ sub request
             {
                 my $nfound = select($rbits, $wbits, undef, $sel_timeout);
                 if ($nfound < 0) {
-                    if ($!{EINTR} || $!{EAGAIN}) {
+                    if ($!{EINTR} || $!{EWOULDBLOCK} || $!{EAGAIN}) {
                         if ($time_before) {
                             $sel_timeout = $sel_timeout_before - (time - $time_before);
                             $sel_timeout = 0 if $sel_timeout < 0;
@@ -291,7 +379,7 @@ sub request
 		my $buf = $socket->_rbuf;
 		my $n = $socket->sysread($buf, 1024, length($buf));
                 unless (defined $n) {
-                    die "read failed: $!" unless  $!{EINTR} || $!{EAGAIN};
+                    die "read failed: $!" unless  $!{EINTR} || $!{EWOULDBLOCK} || $!{EAGAIN};
                     # if we get here the rest of the block will do nothing
                     # and we will retry the read on the next round
                 }
@@ -310,6 +398,7 @@ sub request
 		    if ($code eq "100") {
 			$write_wait = 0;
 			undef($code);
+			goto INITIAL_READ;
 		    }
 		    else {
 			$drop_connection++;
@@ -321,7 +410,7 @@ sub request
 	    if (defined($wbits) && $wbits =~ /[^\0]/) {
 		my $n = $socket->syswrite($$wbuf, length($$wbuf), $woffset);
                 unless (defined $n) {
-                    die "write failed: $!" unless $!{EINTR} || $!{EAGAIN};
+                    die "write failed: $!" unless $!{EINTR} || $!{EWOULDBLOCK} || $!{EAGAIN};
                     $n = 0;  # will retry write on the next round
                 }
                 elsif ($n == 0) {
@@ -378,7 +467,7 @@ sub request
 	{
 	    $n = $socket->read_entity_body($buf, $size);
             unless (defined $n) {
-                redo READ if $!{EINTR} || $!{EAGAIN};
+                redo READ if $!{EINTR} || $!{EWOULDBLOCK} || $!{EAGAIN} || $!{ENOTTY};
                 die "read failed: $!";
             }
 	    redo READ if $n == -1;
@@ -396,13 +485,13 @@ sub request
 
     # keep-alive support
     unless ($drop_connection) {
-	if (my $conn_cache = $self->{ua}{conn_cache}) {
+	if ($cache_key) {
 	    my %connection = map { (lc($_) => 1) }
 		             split(/\s*,\s*/, ($response->header("Connection") || ""));
 	    if (($peer_http_version eq "1.1" && !$connection{close}) ||
 		$connection{"keep-alive"})
 	    {
-		$conn_cache->deposit("http", "$host:$port", $socket);
+		$conn_cache->deposit($self->socket_type, $cache_key, $socket);
 	    }
 	}
     }
@@ -412,44 +501,8 @@ sub request
 
 
 #-----------------------------------------------------------
-package LWP::Protocol::http::SocketMethods;
-
-sub sysread {
-    my $self = shift;
-    if (my $timeout = ${*$self}{io_socket_timeout}) {
-	die "read timeout" unless $self->can_read($timeout);
-    }
-    else {
-	# since we have made the socket non-blocking we
-	# use select to wait for some data to arrive
-	$self->can_read(undef) || die "Assert";
-    }
-    sysread($self, $_[0], $_[1], $_[2] || 0);
-}
-
-sub can_read {
-    my($self, $timeout) = @_;
-    my $fbits = '';
-    vec($fbits, fileno($self), 1) = 1;
-  SELECT:
-    {
-        my $before;
-        $before = time if $timeout;
-        my $nfound = select($fbits, undef, undef, $timeout);
-        if ($nfound < 0) {
-            if ($!{EINTR} || $!{EAGAIN}) {
-                # don't really think EAGAIN can happen here
-                if ($timeout) {
-                    $timeout -= time - $before;
-                    $timeout = 0 if $timeout < 0;
-                }
-                redo SELECT;
-            }
-            die "select failed: $!";
-        }
-        return $nfound > 0;
-    }
-}
+package # hide from PAUSE
+    LWP::Protocol::http::SocketMethods;
 
 sub ping {
     my $self = shift;
@@ -462,8 +515,9 @@ sub increment_response_count {
 }
 
 #-----------------------------------------------------------
-package LWP::Protocol::http::Socket;
-use vars qw(@ISA);
-@ISA = qw(LWP::Protocol::http::SocketMethods Net::HTTP);
+package # hide from PAUSE
+    LWP::Protocol::http::Socket;
+
+use parent -norequire, qw(LWP::Protocol::http::SocketMethods Net::HTTP);
 
 1;
